import logging  

# 创建一个日志记录器  
logger = logging.getLogger()  
logger.setLevel(logging.INFO)  # 设置日志级别为 DEBUG  

# 如果 logger 还没有 handler，则添加
if not logger.handlers:
    # 设置控制台处理器  
    console_handler = logging.StreamHandler()  
    console_handler.setLevel(logging.INFO)  # 控制台日志级别  

    # 设置日志格式  
    formatter = logging.Formatter('%(asctime)s - %(levelname)s:     %(message)s')  
    console_handler.setFormatter(formatter)  

    # 添加处理器到日志记录器  
    logger.addHandler(console_handler)


import google.generativeai as genai
from instructor import Instructor
from pydantic import BaseModel
from typing import Union, Iterator
import instructor
from .base import BaseLLM, register_llm
from chan_agent.logger import logger

@register_llm(model_type="gemini")
class GeminiLLM(BaseLLM):
    """
    GeminiClient is a class that wraps the Google Gemini API.
    """

    def __init__(self, model_name: str = 'gemini-1.5-flash', base_url:str=None, api_key:str='xxx'):
        super().__init__(model_name)
        
        genai.configure(api_key=api_key)
        self.client = genai.GenerativeModel(model_name=self.model_name)
        self.instructor_client: Instructor = instructor.from_gemini(
            self.client,
            mode=instructor.Mode.GEMINI_JSON
        )


    def text_completions(
            self, prompt: str, 
            instructions: str = None, 
            temperature: float = None, 
            top_p: float = None,
            max_tokens: int = None,
            timeout: int = 30,
            return_usage: bool = False,
        ) -> str:
        """
        使用prompt生成文本 completions
        """
        try:
            response = self.client.generate_content(
                contents=prompt,
            )
            return response.text
        except Exception as e:
            logger.error(f"text_completions | Error: {e}")
            raise "error"

    def text_completions_with_messages(
            self, 
            messages: list, 
            temperature: float = None,
            top_p: float = None,
            max_tokens: int = None,
            timeout: int = 30,
            return_usage: bool = False,
        ) -> str:
        """
        使用 messages 列表生成文本 completions。
        """
        try:
            history = []
            for msg in messages:
                if msg['role'] == 'user':
                    history.append({"role": "user", "parts": msg['content']})
                elif msg['role'] == 'assistant':
                    history.append({"role": "model", "parts": msg['content']})
            
            user_input = history[-1]
            history = history[:-1]        
            
            
            chat = self.client.start_chat(
                history=history
            )

            response = chat.send_message(user_input['parts'], stream=False)
            return response.text
        except Exception as e:
            logger.error(f"text_completions_with_messages | Error: {e}")
            return "error"

    def text_completions_with_messages_stream(
            self, 
            messages: list, 
            temperature: float = None,
            top_p: float = None,
            max_tokens: int = None,
            timeout: int = 30,
            return_usage: bool = False,
        ) -> Iterator[str]:
        """
        使用 messages 列表生成文本 completions。
        """

        history = []
        for msg in messages:
            if msg['role'] == 'user':
                history.append({"role": "user", "parts": msg['content']})
            elif msg['role'] == 'assistant':
                history.append({"role": "model", "parts": msg['content']})
        
        user_input = history[-1]
        history = history[:-1]        
        
        
        chat = self.client.start_chat(
            history=history
        )

        # 流式返回部分文本
        full_response = ""
        # 向聊天实例发送消息并处理流式返回
        for chunk in chat.send_message(user_input['parts'], stream=True):
            content = chunk.text
            full_response += content
            yield full_response  # 每次返回更新的部分内容
    
    def basemodel_completions_with_messages(self, basemodel: type[BaseModel], messages: list) -> Union[BaseModel,None]:
        try:
            res = self.instructor_client.chat.completions.create(
                response_model=basemodel,
                messages=messages
            )
            return res
        except Exception as e:
            logger.error(f'basemodel_completions_with_messages | Error: {e}')
            return None



from chan_agent.schema import AgentMessage, ChatMessage, ToolCall
from typing import List, Union
from pydantic import BaseModel
import json
def convert_to_serializable(obj):
    """
    递归地将对象转换为 JSON 可序列化的格式。
    如果对象是 BaseModel 或其他不支持序列化的类型，则将其转换为字典。
    """
    if isinstance(obj, BaseModel):
        return obj.dict()  # 如果是 Pydantic BaseModel，转换为字典
    elif isinstance(obj, list):
        return [convert_to_serializable(item) for item in obj]  # 递归处理列表
    elif isinstance(obj, dict):
        return {key: convert_to_serializable(value) for key, value in obj.items()}  # 递归处理字典
    else:
        return obj  # 直接返回原始值

def messages_dict_to_chat_messages(
        messages: List[dict]
    ) -> List[Union[AgentMessage, ChatMessage]]:

    return [
        AgentMessage(
            role=message['role'],
            content=str(message['content']),
            tool_call=(
                ToolCall.from_str(message['tool_call'])
                if isinstance(message.get('tool_call'), str) else None
            ),            
            extra=message.get('extra', None)
        ) if 'tool_call' in message or 'extra' in message else ChatMessage(
            role=message['role'],
            content=str(message['content'])
        )
        for message in messages
    ]

def chat_messages_to_messages_dict(
        messages: List[Union[AgentMessage, ChatMessage]],
        tool_call_in_content: bool = False,
    ) -> List[dict]:
    return [
        {
            'role': message.role,
            'content': message.content + (f"🛠️{message.tool_call.name}🔚" if tool_call_in_content and isinstance(message, AgentMessage) and message.tool_call else ''),
            **({'tool_call': str(message.tool_call)} if isinstance(message, AgentMessage) and message.tool_call else {}),
            **({
                'extra': convert_to_serializable(message.extra)
            } if isinstance(message, AgentMessage) and message.extra else {}),
        }
        for message in messages
    ]


def get_messages_conversation(
        messages: List[Union[AgentMessage, ChatMessage]], 
        max_content_chat_length:int = 6,
        show_tool_call:bool = True
    ) -> str:
    """
    Get the conversation of the messages.
    """

    num_content_chat_available = max_content_chat_length
    conversation = ""
    for m in messages[::-1]:
        # 倒序获取
        if isinstance(m, AgentMessage):
            conversation = f"[{m.role}]:\n{m.content} {m.tool_call if m.tool_call and show_tool_call else ''}\n\n" + conversation
        else:
            conversation = f"[{m.role}]:\n{m.content}\n\n" + conversation
        if m.role == 'user':
            num_content_chat_available -= 1
            if num_content_chat_available <= 0:
                break

    return conversation
    
def get_messages_conversation_format_openai(
        messages: List[Union[AgentMessage, ChatMessage]], 
        max_content_chat_length:int = 6,
    ) -> List[dict]:
    """
    Get the conversation of the messages.
    """
    conversations = []
    num_content_chat_available = max_content_chat_length
    for m in messages[::-1]:
        # 倒序获取
        if m.role == 'assistant':
            conversations.append(
                {
                    'role': 'assistant',
                    'content': m.content,
                    'function_call': {
                        'name': m.tool_call.name,
                        'arguments': json.dumps(m.tool_call.args),
                    } if isinstance(m, AgentMessage) and m.tool_call else None,
                }
            )
        elif m.role == 'user':
            conversations.append(
                {
                    'role': 'user',
                    'content': m.content,
                }
            )
            num_content_chat_available -= 1
            if num_content_chat_available <= 0:
                break
        elif m.role == 'system':
            conversations.append(
                {
                    'role': 'system',
                    'content': m.content,
                }
            )
        else:
            conversations.append(
                {
                    'role': 'function',
                    'name': m.role,
                    'content': m.content,
                }
            )
    
    return conversations[::-1]


import os
import datetime
import json
from filelock import FileLock 
from pathlib import Path

# 数据库文件路径
logs_dir = Path('logs')

# 确保目录存在
os.makedirs(logs_dir, exist_ok=True)

def llm_log(data: dict):
    today = datetime.date.today()
    log_path = logs_dir / f"llm_logs_{today.isoformat()}.json"
    lockfile = str(log_path) + ".lock"

    # 使用文件锁来避免并发读写冲突
    with FileLock(lockfile):
        if log_path.exists():
            # 读取现有内容
            with open(str(log_path), "r", encoding="utf-8") as f:
                existing_data = json.load(f)
        else:
            existing_data = []

        existing_data.append(data)
        
        # 写入文件
        with open(str(log_path), "w", encoding="utf-8") as f:
            json.dump(existing_data, f, ensure_ascii=False, indent=4)


import time
import datetime
from .db_tracker import llm_log as db_llm_log
from .json_tracker import llm_log as json_llm_log
from chan_agent.config import DEFAULT_LLM_TRACKER

logger_map = {
    "db_tracker": db_llm_log,
    "json_tracker": json_llm_log,
}

def wrap_create(create_fn):
    """
    修饰器：支持openai的接口格式
    """
    logger_fn = logger_map.get(DEFAULT_LLM_TRACKER, None)

    def create(*args, **kwargs):

        # 获取原始输入参数
        user = kwargs.get("user", None)
        functions = kwargs.get("functions", None)
        tools = kwargs.get("tools", None)
        top_p = kwargs.get("top_p", None)
        temperature = kwargs.get("temperature", None)
        input_messages = kwargs.get("messages", None)
        stream = kwargs.get("stream", False)
        max_tokens = kwargs.get("max_tokens", None)
        response_format = kwargs.get("response_format", None)

        wrap_cached = {
            'user': user,
            'stream': stream,
            'top_p': top_p,
            'temperature': temperature,
            'max_tokens': max_tokens,
            'response_format': response_format,
            'functions': functions,
            'tools': tools,
            'input_messages': input_messages,
        }

        start_time = time.perf_counter()
        timestamp = datetime.datetime.now().isoformat()  # 记录调用时间

        if stream:
            # Stream 模式处理逻辑
            response = create_fn(*args, **kwargs)
            outputs_messages = []
            current_message = {"role": None, "content": None, "function_call": None, "tool_calls": None}
            model = None
            usage = None

            def gen():
                nonlocal outputs_messages, current_message, model, usage
                try:
                    for chunk in response:
                        # 提取模型名称
                        if model is None:
                            model = chunk.model

                        usage = {
                            'completion_tokens': chunk.usage.completion_tokens,
                            'prompt_tokens': chunk.usage.prompt_tokens,
                            'total_tokens': chunk.usage.total_tokens
                        } if chunk.usage else None
                        
                        for choice in chunk.choices:
                            delta = choice.delta

                            # 更新 role（通常出现在第一块）
                            if delta.role:
                                if current_message["role"] is None:
                                    if delta.role == "model":
                                        delta.role = "assistant"
                                    current_message["role"] = delta.role

                            # 累积内容
                            if delta.content:
                                if current_message["content"] is None:
                                    current_message["content"] = delta.content
                                else:
                                    current_message["content"] += delta.content

                            
                            # 处理工具调用
                            if delta.function_call:
                                if current_message['function_call'] is None:
                                    current_message['function_call'] = {
                                        "name": delta.function_call.name,
                                        "arguments": delta.function_call.arguments
                                    }
                                else:
                                    current_message['function_call']['arguments'] += delta.function_call.arguments

                            if delta.tool_calls:
                                if current_message['tool_calls'] is None:
                                    current_message['tool_calls'] = [{'name': '', 'arguments': ''} for _ in range(len(delta.tool_calls))]

                                
                                for tool_call in delta.tool_calls:
                                    if tool_call.function:
                                        if tool_call.function.name:
                                            current_message['tool_calls'][tool_call.index]['name'] += tool_call.function.name
                                        if tool_call.function.arguments:
                                            current_message['tool_calls'][tool_call.index]['arguments'] += tool_call.function.arguments
                                    
                            # 完成一条消息
                            if choice.finish_reason in ["stop", "function_call", "tool_calls"]:
                                outputs_messages.append(current_message)
                                current_message = {"role": None, "content": "", "function_call": None, "tool_calls": None}

                        yield chunk
                finally:
                    # 收集最后一条未完成的消息
                    if current_message["content"]:
                        outputs_messages.append(current_message)

                    elapsed_time = time.perf_counter() - start_time
                    wrap_cached.update({
                        "outputs_messages": outputs_messages,
                        "model": model,
                        "usage": usage,
                        "timestamp": timestamp,
                        "elapsed_time": elapsed_time,
                    })
                    
                    if logger_fn is not None:
                        logger_fn(wrap_cached)
                
            return gen()

        else:
            # 非 Stream 模式处理逻辑
            response = create_fn(*args, **kwargs)
            elapsed_time = time.perf_counter() - start_time
            usage = {
                'completion_tokens': response.usage.completion_tokens,
                'prompt_tokens': response.usage.prompt_tokens,
                'total_tokens': response.usage.total_tokens
            } if response.usage else None

            model = response.model if response.model else kwargs.get('model', 'unknown')
            outputs_messages = [
                {
                    'role': i.message.role if i.message.role != 'model' else 'assistant',
                    'content': i.message.content,
                    'tool_calls': [{'name': j.function.name, 'arguments': j.function.arguments} for j in
                                   i.message.tool_calls] if i.message.tool_calls else None,
                    'function_call': {
                        'name': i.message.function_call.name,
                        'arguments': i.message.function_call.arguments
                    } if i.message.function_call else None,
                }
                for i in response.choices
            ]

            wrap_cached.update({
                'outputs_messages': outputs_messages,
                'model': model,
                'usage': usage,
                'timestamp': timestamp,
                'elapsed_time': elapsed_time
            })

            if logger_fn is not None:
                logger_fn(wrap_cached)

            return response

    return create


from abc import ABC
from pydantic import BaseModel
from typing import Union, Iterator, List
from chan_agent.logger import logger
from chan_agent.utils.image import encode_image_from_url

LLM_REGISTRY = {}


def register_llm(model_type):

    def decorator(cls):
        LLM_REGISTRY[model_type] = cls
        return cls

    return decorator




class BaseLLM(ABC):

    def __init__(self, model_name: str = 'gpt-4o-mini'):
        self.model_name = model_name
        
        self.client = None
        self.instructor_client = None

    def set_model_name(self, model_name: str):
        """
        修改模型名称
        """
        self.model_name = model_name

    def image_completions(
            self, 
            prompt: str, 
            images: List[str], 
            instructions: str = None, 
            temperature: float = None,
            top_p: float = None,
            max_tokens: int = None,
            timeout: int = 30,
            return_usage: bool = False
        ) -> str | dict:
        """
        图像分析
        """
        if 'gemini' in self.model_name:
            images = [encode_image_from_url(image_url) for image_url in images]

        messages = []
        if instructions:
            messages.append({"role": "system", "content": instructions})

        # 初始化用户内容列表，包含文本提示
        user_content = [{"type": "text", "text": prompt}]
        # 将图片URLs添加到用户内容列表中
        user_content.extend(
            [{"type": "image_url", "image_url": {"url": img_url}} for img_url in images])
        # 构造消息列表，包括系统指令和用户内容
        messages.append({"role": "user", "content": user_content})
        return self.text_completions_with_messages(messages, temperature=temperature, top_p=top_p, max_tokens=max_tokens, timeout=timeout, return_usage=return_usage)

    def image_basemodel_completions(
            self, 
            basemodel: type[BaseModel], 
            prompt: str, 
            images: List[str],
            instructions: str = None, 
            timeout:int=30
        )  -> Union[BaseModel,None]:
        """
        使用prompt生成basemodel
        """
        if 'gemini' in self.model_name:
            images = [encode_image_from_url(image_url) for image_url in images]

        messages = []
        if instructions:
            messages.append({"role": "system", "content": instructions})

        # 初始化用户内容列表，包含文本提示
        user_content = [{"type": "text", "text": prompt}]
        # 将图片URLs添加到用户内容列表中
        user_content.extend(
            [{"type": "image_url", "image_url": {"url": img_url}} for img_url in images])
        # 构造消息列表，包括系统指令和用户内容
        messages.append({"role": "user", "content": user_content})

        return self.basemodel_completions_with_messages(basemodel, messages, timeout)


    def text_completions_with_messages(
            self, 
            messages: list, 
            temperature: float = None,
            top_p: float = None,
            max_tokens: int = None,
            timeout: int = 30,
            return_usage: bool = False,
        ) -> str | dict:
        """
        使用 messages 列表生成文本 completions。
        """
        try:
            response = self.client.chat.completions.create(
                model=self.model_name,
                messages=messages,
                stream=False,
                temperature=temperature,
                top_p=top_p,
                max_tokens=max_tokens,
                timeout=timeout,
            )
            usage = {
                'completion_tokens': response.usage.completion_tokens,
                'prompt_tokens': response.usage.prompt_tokens,
                'total_tokens': response.usage.total_tokens
            } if response.usage else None
            content = response.choices[0].message.content

            if return_usage:
                return {
                    'content': content,
                    'usage': usage
                }
            else:
                return content
        except Exception as e:
            logger.error(f"text_completions_with_messages | Error: {e}")
            if return_usage:
                return {
                    'content': "error",
                    'usage': None
                }
            else:
                return "error"
    
    def text_completions_with_messages_stream(
            self, 
            messages: list, 
            temperature: float = None,
            top_p: float = None,
            max_tokens: int = None,
            timeout: int = 30,
            return_usage: bool = False,
        ) -> Iterator[Union[str, dict]]:
        """
        使用 messages 列表生成文本 completions。
        """
        try:
            response = self.client.chat.completions.create(
                model=self.model_name,
                messages=messages,
                stream=True,
                temperature=temperature,
                top_p=top_p,
                max_tokens=max_tokens,
                timeout = timeout,
                stream_options={'include_usage': True}
            )

            full_content = ""
            for chunk in response:
                if chunk.choices:
                    choices = chunk.choices[0]
                    if choices.delta.content:
                        full_content += choices.delta.content
                
                if return_usage:
                    usage = {
                        'completion_tokens': chunk.usage.completion_tokens,
                        'prompt_tokens': chunk.usage.prompt_tokens,
                        'total_tokens': chunk.usage.total_tokens
                    } if chunk.usage else None
                    yield {
                        'content': full_content,
                        'usage': usage
                    }
                else:
                    yield full_content                
        except Exception as e:
            logger.error(f"text_completions_with_messages_stream | Error: {e}")
            if return_usage:
                yield {
                    'content': "error",
                    'usage': None
                }
            else:
                yield "error"
        
    
    def text_completions(
            self, 
            prompt: str, 
            instructions: str = None, 
            temperature: float = None, 
            top_p: float = None, 
            max_tokens: int = None, 
            timeout: int = 30,
            return_usage: bool = False,
        ) -> str | dict:
        """
        使用prompt生成文本 completions
        """
        messages = []
        if instructions:
            messages.append({"role": "system", "content": instructions})
        messages.append({"role": "user", "content": prompt})

        return self.text_completions_with_messages(messages, temperature, top_p, max_tokens, timeout, return_usage)
    
    def text_completions_with_stream(
            self, 
            prompt: str, 
            instructions: str = None, 
            temperature: float = None, 
            top_p: float = None, 
            max_tokens: int = None, 
            timeout: int = 30,
            return_usage: bool = False
        )-> Iterator[Union[str, dict]]:
        """
        使用prompt生成文本 completions 流式返回
        """
        messages = []
        if instructions:
            messages.append({"role": "system", "content": instructions})
        messages.append({"role": "user", "content": prompt})
        return self.text_completions_with_messages_stream(messages, temperature, top_p, max_tokens, timeout, return_usage)
        
    
    def basemodel_completions(self, basemodel: type[BaseModel], prompt: str, instructions: str = None, timeout:int=30)  -> Union[BaseModel,None]:
        """
        使用prompt生成basemodel
        """
        messages = [{"role": "user", "content": prompt}]
        if instructions:
            messages.append({"role": "system", "content": instructions})

        return self.basemodel_completions_with_messages(basemodel, messages, timeout)

    def basemodel_completions_with_messages(self, basemodel: type[BaseModel], messages: list, timeout:int=30) -> Union[BaseModel,None]:
        """
        使用messages列表生成basemodel
        """
        # BUG 这里可能回卡死
        try:
            res = self.instructor_client.chat.completions.create(
                model=self.model_name,
                response_model=basemodel,
                messages=messages,
                max_retries=3,
                timeout=timeout  # 内部timeout参数保留，但主要依赖外部超时
            )
            return res
        except TimeoutError:
            logger.error(f'basemodel_completions_with_messages | Timeout after {timeout} seconds')
            return None
        except Exception as e:
            logger.error(f'basemodel_completions_with_messages | Internal Error: {e}')
            return None

    

        
        


from functools import cache
from typing import Any
from openai import OpenAI
import instructor
from .base import BaseLLM, register_llm
from chan_agent.llm_track import wrap_create
from typing import Any, Callable

class OpenAICredentialsRefresher:
    def __init__(self, token_fetcher: Callable[[], str], **kwargs: Any) -> None:
        """
        OpenAI 客户端凭据刷新器

        :param token_fetcher: 获取访问令牌的外部方法
        """
        self.token_fetcher = token_fetcher  # 传递的外部方法
        self.client = OpenAI(**kwargs, api_key="DUMMY")

    def __getattr__(self, name: str) -> Any:
        """
        动态代理 OpenAI 客户端的方法，并在需要时刷新 Token
        """
        new_token = self.token_fetcher()  # 获取新的 Token
        if not new_token:
            raise RuntimeError("Unable to refresh auth token")

        self.client.api_key = new_token
        return getattr(self.client, name)

@cache
def init_openai_client(
    project_id: str = "project_id",
    location: str = "asia-east1",
    endpoint_id: str = 'openapi',
    token_fetcher: Callable[[], str] = None  # 额外的参数
):
    """
    初始化 OpenAI 客户端
    """
    if token_fetcher is None:
        raise ValueError("token_fetcher function must be provided")

    # 初始化 OpenAI 客户端，并传递 token_fetcher
    client = OpenAICredentialsRefresher(
        token_fetcher=token_fetcher,
        base_url=f"https://{location}-aiplatform.googleapis.com/v1beta1/projects/{project_id}/locations/{location}/endpoints/{endpoint_id}",
    )

    client.chat.completions.create = wrap_create(create_fn=client.chat.completions.create)
    return client

@register_llm(model_type='vertexai')
class VertexLLM(BaseLLM):
    def __init__(
            self, 
            model_name: str = 'google/gemini-1.5-flash-002',
            base_url: str = None,
            api_key: str = None, 
            **kwargs
        ):
        super().__init__(model_name)
        project_id = kwargs.get('project_id')
        loccation = kwargs.get('location')
        token_fetcher = kwargs.get('token_fetcher')
        endpoint_id = kwargs.get('endpoint_id', 'openapi')
        self.client = init_openai_client(project_id, loccation, endpoint_id, token_fetcher)
     
        self.instructor_client = instructor.from_openai(
            self.client.client,
            mode=instructor.Mode.JSON
        )
        
    


    





from .llms import BaseLLM
from .schema import TaskOutputs, TaskInputItem
from typing import List, Union
from pydantic import BaseModel
from textwrap import dedent


PROMPT_TASK_BACKBONE = dedent("""\
# Task 
{task}
                               
# Outputs Format
{outputs}

{rules}
                              
# Inputs
Please reference the following data and adhere to the rules as much as possible to complete the task:
{inputs}
""")

class TaskLLM:

    def __init__(
        self, 
        llm: BaseLLM, 
        task: str,
        rules: List[str], 
        output_model: TaskOutputs,
    ):
        self.llm = llm
        self.task = task
        self.rules = ""
        if len(rules):
            self.rules = "# Rules\n"
            for rule in rules:
                self.rules += f"- {rule}\n"
        self.output_model = output_model

    def call(self, inputs: List[TaskInputItem], **kwargs) -> Union[BaseModel, None]:
        
        prompt = PROMPT_TASK_BACKBONE.format(
            task = self.task,
            inputs = "\n\n".join([str(ip) for ip in inputs]),
            outputs = self.output_model.get_scheme(),
            rules = self.rules
        )

        return self.llm.basemodel_completions(self.output_model, prompt)
    


class ImageTaskLLM(TaskLLM):

    def call(self, inputs: List[TaskInputItem], images: List[str], **kwargs):

        inputs.append(
            TaskInputItem(
                key="images",
                key_name="images",
                value="See attached images",
            ))
        
        prompt = PROMPT_TASK_BACKBONE.format(
            task = self.task,
            inputs = "\n\n".join([str(ip) for ip in inputs]),
            outputs = self.output_model.get_scheme(),
            rules = self.rules
        )
        
        return self.llm.image_basemodel_completions(
            self.output_model,
            prompt = prompt,
            images = images
        )

        



from .base import BaseLLM, LLM_REGISTRY
from .openai import OpenaiLLM
from .gemini import GeminiLLM
from .vertex import VertexLLM
from functools import cache


@cache
def get_llm(llm_type:str, model_name:str,  base_url:str, api_key:str, **kwargs) -> BaseLLM:
    """
    获取llm
    """
    if llm_type not in LLM_REGISTRY:
        raise ValueError(f"Invalid llm_type: {llm_type}. Must be one of {list(LLM_REGISTRY.keys())}.")
   
    return LLM_REGISTRY[llm_type](model_name, base_url, api_key, **kwargs)


__all__ = [
    'BaseLLM',
    'LLM_REGISTRY',
    'get_llm',
]


import os
from playhouse.sqlite_ext import *
from datetime import datetime
from chan_agent.logger import logger

# 数据库文件路径
db_path = 'logs/db.sqlite'

# 确保目录存在
os.makedirs(os.path.dirname(db_path), exist_ok=True)

db = SqliteDatabase(db_path)


class LLMTracker(Model):
    id = AutoField()        # 自动递增主键
    data = JSONField()      # JSON 格式的字段
    created_at = DateTimeField(default=datetime.now)  # 自动记录创建时间

    class Meta:
        database = db
        table_name = 'llm_tracker'

    @classmethod
    def initialize_table(cls):
        """
        检查表是否存在，不存在时自动创建。
        """
        if not db.table_exists(cls._meta.table_name):
            db.create_tables([cls])
            logger.info(f"Table '{cls._meta.table_name}' created successfully.")

    @classmethod
    def create_log(cls, data: dict):
        """
        写入一条新数据到 LLMTracker 表。
        
        Args:
            data (dict): 字典数据。

        Returns:
            LLMTracker: 返回插入的记录对象。
        """
        try:
            cls.initialize_table()
            with db.atomic():  # 启用事务
                new_record = cls.create(data=data)
            return new_record
        except Exception as e:
            raise ValueError(f"Failed to write data to LLMTracker: {e}")
    

def llm_log(data: dict):
    LLMTracker.create_log(data)



from functools import cache
from openai import OpenAI
import instructor
from .base import BaseLLM, register_llm
from chan_agent.llm_track import wrap_create
from chan_agent.logger import logger

@cache
def init_openai_client(base_url:str, api_key:str,**kwargs):
    """
    初始化client客户端
    """
    # 定义openai client
    client = OpenAI(
        base_url=base_url,
        api_key=api_key,
        **kwargs
    )

    client.chat.completions.create = wrap_create(create_fn=client.chat.completions.create)

    return client

@register_llm(model_type="openai")
class OpenaiLLM(BaseLLM):
    def __init__(self, model_name: str = 'gpt-4o-mini', base_url:str=None, api_key:str='xxx'):
        super().__init__(model_name)

        self.client = init_openai_client(base_url = base_url, api_key = api_key)
        if self.model_name is not None and self.model_name.startswith("gpt"):
            instructor_mode = instructor.Mode.TOOLS    
        else:
            # 兼任ollama等openai接口模型的其他模型
            instructor_mode = instructor.Mode.JSON
        self.instructor_client = instructor.from_openai(
            self.client,
            mode=instructor_mode
        )

    def fc_completions_with_messages(
        self, 
        messages: list, 
        functions: list,
        temperature: float = None,
        top_p: float = None,
        max_tokens: int = None,
        timeout: int = 30,
        return_usage: bool = False,
    ):
        try:
            response = self.client.chat.completions.create(
                model=self.model_name,
                messages=messages,
                functions=functions,
                stream=False,
                temperature=temperature,
                top_p=top_p,
                max_tokens=max_tokens,
                timeout=timeout,
            )
            usage = {
                'completion_tokens': response.usage.completion_tokens,
                'prompt_tokens': response.usage.prompt_tokens,
                'total_tokens': response.usage.total_tokens
            } if response.usage else None
            content = response.choices[0].message.content
            function_call = None
            if response.choices[0].message.function_call:
                function_call = {
                    "name": response.choices[0].message.function_call.name,
                    "arguments": response.choices[0].message.function_call.arguments
                }

            if return_usage:
                return {
                    'content': content,
                    'function_call': function_call,
                    'usage': usage
                }
            else:
                return {
                    'content': content,
                    'function_call': function_call
                }
        except Exception as e:
            logger.error(f"text_completions_with_messages | Error: {e}")
            if return_usage:
                return {
                    'content': "error",
                    'function_call': None,
                    'usage': None
                }
            else:
                return {
                    'content': "error",
                    'function_call': None
                }

    def fc_completions_with_messages_stream(
        self, 
        messages: list, 
        functions: list,
        temperature: float = None,
        top_p: float = None,
        max_tokens: int = None,
        timeout: int = 30,
        return_usage: bool = False,
    ):
        
        try:
            response = self.client.chat.completions.create(
                model=self.model_name,
                messages=messages,
                functions=functions,
                stream=True,
                temperature=temperature,
                top_p=top_p,
                max_tokens=max_tokens,
                timeout = timeout,
                stream_options={'include_usage': True}
            )

            full_content = ""
            function_call = None
            function_call_finish = False
            for chunk in response:
                if chunk.choices:
                    choices = chunk.choices[0]
                    if choices.delta.content:
                        full_content += choices.delta.content
                    if choices.delta.function_call:
                        if function_call is None:
                            function_call = {
                                "name": choices.delta.function_call.name,
                                "arguments": choices.delta.function_call.arguments
                            }
                        else:
                            function_call["arguments"] += choices.delta.function_call.arguments
                    if choices.finish_reason == 'function_call':
                        function_call_finish = True

                
                if return_usage:
                    usage = {
                        'completion_tokens': chunk.usage.completion_tokens,
                        'prompt_tokens': chunk.usage.prompt_tokens,
                        'total_tokens': chunk.usage.total_tokens
                    } if chunk.usage else None
                    yield {
                        'content': full_content,
                        'function_call': function_call if function_call_finish else None,
                        'usage': usage
                    }
                else:
                    yield {
                        'content': full_content,
                        'function_call': function_call if function_call_finish else None
                    }                
        except Exception as e:
            logger.error(f"fc_completions_with_messages_stream | Error: {e}")
            if return_usage:
                yield {
                    'content': "error",
                    'function_call': None,
                    'usage': None
                }
            else:
                yield {
                    'content': 'error',
                    'function_call': None
                } 



import json
from abc import ABC, abstractmethod
from typing import List, Optional, Union
from .logger import logger
from .schema import ToolResult

TOOL_REGISTRY = {}


def register_tool(name, allow_overwrite=False):

    def decorator(cls):
        if name in TOOL_REGISTRY:
            if allow_overwrite:
                logger.warning(f'Tool `{name}` already exists! Overwriting with class {cls}.')
            else:
                raise ValueError(f'Tool `{name}` already exists! Please ensure that the tool name is unique.')
        if cls.name and (cls.name != name):
            raise ValueError(f'{cls.__name__}.name="{cls.name}" conflicts with @register_tool(name="{name}").')
        cls.name = name
        TOOL_REGISTRY[name] = cls

        return cls

    return decorator


def schema_to_str(schema: Union[List[dict], dict], indent_level=0):
    """
    转换为字符串描述
    """
    output = []
    
    # 设置缩进
    indent = "  " * indent_level

    for key in schema:
        type_ = schema[key].get('type')
        description = schema[key].get('description', '')
        items = schema[key].get('items', {})
        properties = schema[key].get('properties', {})
        required = schema[key].get('required', True)

        if type_ == 'array' and items:
            # 数组
            item_type = items.get('type')
            item_properties = items.get('properties', {})
            if item_type == 'object' and item_properties:
                output.append(f"{indent}- `{key}`: list[object] {description}, each object with the following properties:")
                output.extend(schema_to_str(item_properties, indent_level + 1))
            else:
                output.append(f"{indent}- `{key}`: list[{item_type}], {description}")

        elif type_ == 'object' and properties:
            # 如果是对象
            output.append(f"{indent}- `{key}`: object {description}, with the following properties:")
            output.extend(schema_to_str(properties, indent_level + 1))
        else:
            output.append(f"{indent}- `{key}`: {type_ if required else f'Optional[{type_}]'} {description}")
        
    return output

    

class BaseTool(ABC):
    name: str = ''
    description: str = ''
    parameters: List[dict] = {
        'your_key_name': {
            'type': 'array',
            'items': {
                'type': 'number',
            },
            'description': 'your key description',
            'required': True
        },
    }

    def __init__(self, cfg: Optional[dict] = None):
        self.cfg = cfg or {}
        if not self.name:
            raise ValueError(
                f'You must set {self.__class__.__name__}.name, either by @register_tool(name=...) or explicitly setting {self.__class__.__name__}.name'
            )

    def __str__(self):
        output = []
        
        # 名称和描述
        output.append(f"- {self.name}: {self.description}, with the following input parameters:")

        output.extend(schema_to_str(self.parameters, 1))

        return '\n'.join(output)
    
    def to_function(self) -> dict:
        """
        Convert the tool to a function format compatible with OpenAI Function Calling.
        """
        def process_schema(schema):
            """
            递归处理 JSON Schema，提取 required 字段，并转换格式。
            """
            if not isinstance(schema, dict):
                return schema
            
            properties = {}
            required_fields = []
            
            for key, value in schema.items():
                if isinstance(value, dict):
                    new_value = value.copy()
                    
                    if "items" in value and value['items']['type'] == 'object':
                        new_value['items'] = process_schema(value['items']['properties'])

                    # 递归处理嵌套结构（如 properties）
                    if "properties" in value:
                        new_value["properties"] = process_schema(value["properties"])

                    # 提取 required 信息并更新属性
                    if "required" in value and value["required"]:
                        required_fields.append(key)
                    
                    # 删除嵌套结构中的 required 字段
                    new_value.pop("required", None)

                    properties[key] = new_value
            
            # 构建最终的 schema
            output_schema = {"type": "object", "properties": properties}
            if required_fields:
                output_schema["required"] = required_fields

            return output_schema

        # 处理并格式化参数 schema
        processed_parameters = process_schema(self.parameters)
        
        return {
            "name": self.name,
            "description": self.description or "No description available.",
            "parameters": processed_parameters
        }
        
        

    @abstractmethod
    def call(self, params: Union[str, dict], **kwargs) -> ToolResult:
        """The interface for calling tools.

        Each tool needs to implement this function, which is the workflow of the tool.

        Args:
            params: The parameters of func_call.
            kwargs: Additional parameters for calling tools.

        Returns:
            The result returned by the tool, implemented in the subclass.
        """
        raise NotImplementedError

    def _verify_json_format_args(self, params: Union[str, dict]) -> dict:
        """Verify the parameters of the function call"""
        if isinstance(params, str):
            try:
                params_json: dict = json.loads(params)
            except json.decoder.JSONDecodeError:
                raise ValueError('Parameters must be formatted as a valid JSON!')
        else:
            params_json: dict = params
        
        # TODO 验证参数格式

        return params_json




import requests
import base64

# 下载图像并转为base64
def encode_image_from_url(image_url):
    # 下载图像
    response = requests.get(image_url)
    if response.status_code == 200:
        # 将图像内容转为base64
        return "data:image/jpeg;base64," + base64.b64encode(response.content).decode('utf-8')
    else:
        return None


from setuptools import setup, find_packages

setup(
    name="chan-agent",  # 包的名称
    version="0.0.13",    # 版本号
    author="Chan",
    author_email="925355568@qq.com",
    description="A simple llm agent",
    long_description=open("README.md").read(),
    long_description_content_type="text/markdown",
    url="https://github.com/Chan-0312/chan-agent",  # 项目主页
    packages=find_packages(),  # 自动发现包
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires=">=3.11",  # Python 的最低版本要求
    install_requires=[        # 运行时依赖的包
        "openai",
        "instructor",
        "google-generativeai",
        "filelock",
        "peewee"
    ],
)



import json
from .logger import logger
from textwrap import dedent
from .llms import BaseLLM
from .base_tool import BaseTool
from .schema import *
from .utils.messages_processing import get_messages_conversation_format_openai
from typing import List, Union, Optional, Iterator, Literal


PROMPT_ROLE_SECTION = dedent("""\
# Role
{role}

""")


PROMPT_RULES_SECTION = dedent("""\
# Rules
- Maintain a polite and professional tone at all times
{rules}

{extra_info}

""")


class FCAgent:

    def __init__(
        self, 
        llm: BaseLLM,
        role: str, 
        tools: List[BaseTool],
        rules: List[str],
        static_extra_info: Optional[str] = "",
        agent_type: Literal["user", "assistant"] = "assistant",
        max_content_chat_length: int = 6,
        max_llm_call_per_run: int = 3,
    ):

        self.role = role
        self.tools_map = {tool.name: tool for tool in tools}
        self.tools_desc = "\n".join([str(tool) for tool in tools])
        self.rules_desc = "\n".join([f"- {rule}" for rule in rules])
        self.static_extra_info = static_extra_info

        # 最大每一轮工具使用的轮数
        self.max_llm_call_per_run = max_llm_call_per_run
        # 最大对话上下文的轮数
        self.max_content_chat_length = max_content_chat_length
        # agent类型
        self.agent_type = agent_type

        self.llm = llm
        self.usage = {
            "prompt_tokens": 0,
            "completion_tokens": 0,
            "total_tokens": 0,
        }

    def record_usage(self, usage: dict):
        """
        缓存usage
        """
        if usage:
            self.usage["total_tokens"] += usage.get("total_tokens", 0)
            self.usage["prompt_tokens"] += usage.get("prompt_tokens", 0)
            self.usage["completion_tokens"] += usage.get("completion_tokens", 0)
    
    def clear_usage(self):
        """
        清除usage
        """
        self.usage = {
            "prompt_tokens": 0,
            "completion_tokens": 0,
            "total_tokens": 0,
        }

    def get_usage(self):
        """
        获取usage
        """
        return self.usage


    def make_agent_messages(self, messages:List[AgentMessage], **kwargs) -> str:
        """
        构建agent messages
        """

        # 仅保留max_content_chat_length长度的上下文
        conversations = get_messages_conversation_format_openai(messages, max_content_chat_length=self.max_content_chat_length)

        system_prompt = ""
        system_prompt += PROMPT_ROLE_SECTION.format(role=self.role)
        system_prompt += PROMPT_RULES_SECTION.format(
            rules=self.rules_desc, 
            extra_info=self.static_extra_info + kwargs.get("dynamic_extra_info", "")
        )

        return [{'role': 'system', 'content': system_prompt}] + conversations
    
    def __detect_tool(self, content:dict) -> Union[ToolCall, SystemError, None]:
        """
        检测工具
        """
        try:
            tool_call = ToolCall(name = content['function_call'].get("name"), args = json.loads(content['function_call'].get("arguments", {}))) if content.get('function_call') else None
            return tool_call
        except Exception as e:
            return SystemError(error=str(e))

    def chat(self, messages:List[AgentMessage], do_tool_call:bool=True, **kwargs) -> List[AgentMessage]:
        """
        对话
        """
        # 清除usage
        self.clear_usage()

        num_llm_calls_available = self.max_llm_call_per_run
        response = []
        while True and num_llm_calls_available > 0:
            system_error = None

            num_llm_calls_available -= 1

            # 生成agent prompt
            agent_messages = self.make_agent_messages(messages=messages+response, **kwargs)

            # 发起llm请求
            out_dict = self.llm.fc_completions_with_messages(
                messages=agent_messages,
                functions = [tool.to_function() for tool in self.tools_map.values()],
                temperature=kwargs.get('temperature', 0.3),
                top_p=kwargs.get('top_p', None),
                max_tokens=kwargs.get('max_tokens', None),
                return_usage=True
            )

            # 记录usage
            self.record_usage(out_dict.get('usage', {}))

            # 检测使用的工具
            tool_call = self.__detect_tool(out_dict)

            response.append(AgentMessage(
                role=self.agent_type,
                content=out_dict['content'] or '',
                tool_call=None if isinstance(tool_call, SystemError) else tool_call,
            ))
            if isinstance(tool_call, SystemError):
                system_error = tool_call
                

            if do_tool_call and isinstance(tool_call, ToolCall):
                # 执行工具
                logger.info(tool_call)
                tool_result = self.tools_map[tool_call.name].call(params=tool_call.args, **kwargs)
                response.append(AgentMessage(
                    role=tool_call.name,
                    content=tool_result.response,
                    extra=tool_result.extra,
                ))
                if tool_result.use_tool_response:
                    response.append(AgentMessage(
                        role=self.agent_type,
                        content=tool_result.response,
                    ))
            elif system_error:
                response.append(AgentMessage(
                    role="system",
                    content=tool_call.error,
                ))
            else:
                break 
                
        return response


    def chat_with_stream(self, messages:List[AgentMessage], do_tool_call:bool=True, **kwargs) -> Iterator[List[AgentMessage]]:
        """
        流式对话
        """
        # 清除usage
        self.clear_usage()
        num_llm_calls_available = self.max_llm_call_per_run
        response = []
        while True and num_llm_calls_available > 0:
            system_error = None
            
            num_llm_calls_available -= 1

            # 生成agent prompt
            agent_messages = self.make_agent_messages(messages=messages+response, **kwargs)

            # 发起llm请求
            llm_out = self.llm.fc_completions_with_messages_stream(
                messages=agent_messages, 
                functions=[tool.to_function() for tool in self.tools_map.values()],
                temperature=kwargs.get('temperature', 0.3),
                top_p=kwargs.get('top_p', None),
                max_tokens=kwargs.get('max_tokens', None),
                return_usage=True
            )

            agent_output = AgentMessage(role=self.agent_type, content='')
            for out_dict in llm_out:
                # 检测使用的工具
                tool_call = self.__detect_tool(out_dict)
                agent_output.content = out_dict['content'] or ''
                agent_output.tool_call = None if isinstance(tool_call, SystemError) else tool_call
                self.record_usage(out_dict.get('usage', {}))

                yield response + [agent_output]

                if tool_call is not None:
                    # 已经识别到工具直接截断
                    if isinstance(tool_call, SystemError):
                        system_error = tool_call
                    # break
            
            response += [agent_output]
            
            if do_tool_call and isinstance(agent_output.tool_call, ToolCall):
                # 执行工具
                logger.info(agent_output.tool_call)

                tool_result = self.tools_map[tool_call.name].call(params=tool_call.args, **kwargs)
                response.append(AgentMessage(
                    role=tool_call.name,
                    content=tool_result.response,
                    extra=tool_result.extra,
                ))
                yield response
                if tool_result.use_tool_response:
                    # 使用工具的回复
                    response.append(AgentMessage(
                        role=self.agent_type,
                        content=tool_result.response,
                    ))
                    yield response
                    break
            elif system_error:
                response.append(AgentMessage(
                    role="system",
                    content=system_error.error,
                ))
                yield response
            else:
                break 
                        
        


from pydantic import BaseModel, Field
from typing import List, Optional, Any
import json

class ToolCall(BaseModel):
    name: str = Field(..., description="tool name")
    args: Any = Field(None, description="tool args")

    def __str__(self):
        return f'🛠️ {self.model_dump_json()} 🔚'
    
    @classmethod
    def from_str(cls, formatted_str: str):
        """
        Initializes a ToolCall object from a formatted string in the format:
        🛠️ {json_data} 🔚. Returns None if parsing fails.
        """
        try:
            # Split the string using delimiters
            parts = formatted_str.split('🛠️ ', 1)
            if len(parts) != 2 or ' 🔚' not in parts[1]:
                return None  # Return None if format doesn't match
            
            json_data = parts[1].rsplit(' 🔚', 1)[0].strip()  # Extract the JSON part
            
            # Attempt to parse JSON
            data = json.loads(json_data)
            return cls(**data)
        except (ValueError, json.JSONDecodeError):
            return None  # Return None if any error occurs
    
class ToolResult(BaseModel):
    response: str = Field(..., description="tool content result")
    use_tool_response: bool = Field(False, description="use tool response")
    extra: Any = Field(None, description="extra info")

class ChatMessage(BaseModel):
    role: str = Field(..., description="role")
    content: str = Field(..., description="content")


class AgentMessage(ChatMessage):
    tool_call: Optional[ToolCall] = Field(None, description="tool calls")
    extra: Optional[Any] = Field(None, description="extra info")

class SystemError(BaseModel):
    error: str = Field(..., description="error message")    

class TaskInputItem(BaseModel):
    key: str = Field(..., description="key")
    key_name: str = Field(..., description="key name")
    value: str = Field(..., description="value")
    format_as_code: bool = Field(False, description="Format value as code block if true")

    def __str__(self) -> str:
        if self.format_as_code:
            return f"- **{self.key_name}**:\n```\n{self.value}```"
        return f"- **{self.key_name}**: {self.value}"
    

class TaskOutputs(BaseModel):

    @classmethod
    def get_scheme(cls) -> str:
        return str(cls.model_json_schema())